from enum import StrEnum


class MinecraftBiomeTypes(StrEnum):
    BambooJungle = "minecraft:bamboo_jungle"
    BambooJungleHills = "minecraft:bamboo_jungle_hills"
    BasaltDeltas = "minecraft:basalt_deltas"
    Beach = "minecraft:beach"
    BirchForest = "minecraft:birch_forest"
    BirchForestHills = "minecraft:birch_forest_hills"
    BirchForestHillsMutated = "minecraft:birch_forest_hills_mutated"
    BirchForestMutated = "minecraft:birch_forest_mutated"
    CherryGrove = "minecraft:cherry_grove"
    ColdBeach = "minecraft:cold_beach"
    ColdOcean = "minecraft:cold_ocean"
    ColdTaiga = "minecraft:cold_taiga"
    ColdTaigaHills = "minecraft:cold_taiga_hills"
    ColdTaigaMutated = "minecraft:cold_taiga_mutated"
    CrimsonForest = "minecraft:crimson_forest"
    DeepColdOcean = "minecraft:deep_cold_ocean"
    DeepDark = "minecraft:deep_dark"
    DeepFrozenOcean = "minecraft:deep_frozen_ocean"
    DeepLukewarmOcean = "minecraft:deep_lukewarm_ocean"
    DeepOcean = "minecraft:deep_ocean"
    DeepWarmOcean = "minecraft:deep_warm_ocean"
    Desert = "minecraft:desert"
    DesertHills = "minecraft:desert_hills"
    DesertMutated = "minecraft:desert_mutated"
    DripstoneCaves = "minecraft:dripstone_caves"
    ExtremeHills = "minecraft:extreme_hills"
    ExtremeHillsEdge = "minecraft:extreme_hills_edge"
    ExtremeHillsMutated = "minecraft:extreme_hills_mutated"
    ExtremeHillsPlusTrees = "minecraft:extreme_hills_plus_trees"
    ExtremeHillsPlusTreesMutated = "minecraft:extreme_hills_plus_trees_mutated"
    FlowerForest = "minecraft:flower_forest"
    Forest = "minecraft:forest"
    ForestHills = "minecraft:forest_hills"
    FrozenOcean = "minecraft:frozen_ocean"
    FrozenPeaks = "minecraft:frozen_peaks"
    FrozenRiver = "minecraft:frozen_river"
    Grove = "minecraft:grove"
    Hell = "minecraft:hell"
    IceMountains = "minecraft:ice_mountains"
    IcePlains = "minecraft:ice_plains"
    IcePlainsSpikes = "minecraft:ice_plains_spikes"
    JaggedPeaks = "minecraft:jagged_peaks"
    Jungle = "minecraft:jungle"
    JungleEdge = "minecraft:jungle_edge"
    JungleEdgeMutated = "minecraft:jungle_edge_mutated"
    JungleHills = "minecraft:jungle_hills"
    JungleMutated = "minecraft:jungle_mutated"
    LegacyFrozenOcean = "minecraft:legacy_frozen_ocean"
    LukewarmOcean = "minecraft:lukewarm_ocean"
    LushCaves = "minecraft:lush_caves"
    MangroveSwamp = "minecraft:mangrove_swamp"
    Meadow = "minecraft:meadow"
    MegaTaiga = "minecraft:mega_taiga"
    MegaTaigaHills = "minecraft:mega_taiga_hills"
    Mesa = "minecraft:mesa"
    MesaBryce = "minecraft:mesa_bryce"
    MesaPlateau = "minecraft:mesa_plateau"
    MesaPlateauMutated = "minecraft:mesa_plateau_mutated"
    MesaPlateauStone = "minecraft:mesa_plateau_stone"
    MesaPlateauStoneMutated = "minecraft:mesa_plateau_stone_mutated"
    MushroomIsland = "minecraft:mushroom_island"
    MushroomIslandShore = "minecraft:mushroom_island_shore"
    Ocean = "minecraft:ocean"
    PaleGarden = "minecraft:pale_garden"
    Plains = "minecraft:plains"
    RedwoodTaigaHillsMutated = "minecraft:redwood_taiga_hills_mutated"
    RedwoodTaigaMutated = "minecraft:redwood_taiga_mutated"
    River = "minecraft:river"
    RoofedForest = "minecraft:roofed_forest"
    RoofedForestMutated = "minecraft:roofed_forest_mutated"
    Savanna = "minecraft:savanna"
    SavannaMutated = "minecraft:savanna_mutated"
    SavannaPlateau = "minecraft:savanna_plateau"
    SavannaPlateauMutated = "minecraft:savanna_plateau_mutated"
    SnowySlopes = "minecraft:snowy_slopes"
    SoulsandValley = "minecraft:soulsand_valley"
    StoneBeach = "minecraft:stone_beach"
    StonyPeaks = "minecraft:stony_peaks"
    SunflowerPlains = "minecraft:sunflower_plains"
    Swampland = "minecraft:swampland"
    SwamplandMutated = "minecraft:swampland_mutated"
    Taiga = "minecraft:taiga"
    TaigaHills = "minecraft:taiga_hills"
    TaigaMutated = "minecraft:taiga_mutated"
    TheEnd = "minecraft:the_end"
    WarmOcean = "minecraft:warm_ocean"
    WarpedForest = "minecraft:warped_forest"


class MinecraftBiomeTags(StrEnum):
    Animal = "animal"
    BasaltDeltas = "basalt_deltas"
    Beach = "beach"
    BeeHabitat = "bee_habitat"
    Birch = "birch"
    Caves = "caves"
    CherryGrove = "cherry_grove"
    Cold = "cold"
    CrimsonForest = "crimson_forest"
    DarkOak = "dark_oak"
    Deep = "deep"
    DeepDark = "deep_dark"
    Desert = "desert"
    DripstoneCaves = "dripstone_caves"
    Edge = "edge"
    ExtremeHills = "extreme_hills"
    FastFishing = "fast_fishing"
    FlowerForest = "flower_forest"
    Forest = "forest"
    Frozen = "frozen"
    FrozenPeaks = "frozen_peaks"
    Grove = "grove"
    HighSeas = "high_seas"
    Hills = "hills"
    Ice = "ice"
    IcePlains = "ice_plains"
    JaggedPeaks = "jagged_peaks"
    Jungle = "jungle"
    Lakes = "lakes"
    Lukewarm = "lukewarm"
    LushCaves = "lush_caves"
    MangroveSwamp = "mangrove_swamp"
    Mega = "mega"
    Mesa = "mesa"
    Monster = "monster"
    MooshroomIsland = "mooshroom_island"
    Mountain = "mountain"
    Mutated = "mutated"
    Nether = "nether"
    NetherWastes = "nether_wastes"
    NetherwartForest = "netherwart_forest"
    NoLegacyWorldgen = "no_legacy_worldgen"
    Ocean = "ocean"
    Overworld = "overworld"
    OverworldGeneration = "overworld_generation"
    Plains = "plains"
    Plateau = "plateau"
    Rare = "rare"
    River = "river"
    Roofed = "roofed"
    Savanna = "savanna"
    Shore = "shore"
    Slime = "slime"
    SnowySlopes = "snowy_slopes"
    SoulsandValley = "soulsand_valley"
    SpawnEndermen = "spawn_endermen"
    SpawnFewPiglins = "spawn_few_piglins"
    SpawnFewZombifiedPiglins = "spawn_few_zombified_piglins"
    SpawnGhast = "spawn_ghast"
    SpawnMagmaCubes = "spawn_magma_cubes"
    SpawnManyMagmaCubes = "spawn_many_magma_cubes"
    SpawnPiglin = "spawn_piglin"
    SpawnZombifiedPiglin = "spawn_zombified_piglin"
    Stone = "stone"
    SurfaceMineshaft = "surface_mineshaft"
    Swamp = "swamp"
    SwampWaterHugeMushroom = "swamp_water_huge_mushroom"
    Taiga = "taiga"
    TheEnd = "the_end"
    Warm = "warm"
    WarpedForest = "warped_forest"
